# from fasta
def parse_fasta(filename,limit=-1):
  '''function to parse fasta'''
  header = []
  sequence = []
  lines = open(filename, "r")
  for line in lines:
    line = line.rstrip()
    if line[0] == ">":
      if len(header) == limit:
        break
      header.append(line[1:])
      sequence.append([])
    else:
      sequence[-1].append(line)
  lines.close()
  sequence = [''.join(seq) for seq in sequence]
  return np.array(sequence)

# accuracy of long-range contacts
# (Cb-Cb distance < 8A)
def cont_acc(pred, dist, rset, frac=1.0):

    nc = dist.shape[0]

    # if too short, then there are no
    # long-range contacts
    if pred.shape[0] < 24:
        return 0.0

    # convert distogram to contacts
    w = np.sum(pred[:,:,:12],axis=-1) #sum probs < 8.0A

    # extract indices of the lower tringle
    # along with predicted probabilities
    idx = np.array([[i,j,w[i,j]] for i in range(nc) for j in range(i+1,nc) if rset[j]-rset[i]>=24])


    # select top contacts and check
    # how many are realized in the structure
    n = int(nc*frac)
    #print(w.shape, dist.shape, idx.shape, n, rset)
    top = idx[np.argsort(idx[:,2])][-n:, :2].astype(int)
    ngood = np.sum([dist[r[0],r[1]]<8.0 for r in top])

    return 1.0*ngood/n

# calculate dihedral angles defined by 4 sets of points
def get_dihedrals(a, b, c, d):

    b0 = -1.0*(b - a)
    b1 = c - b
    b2 = d - c

    b1 /= np.linalg.norm(b1, axis=-1)[:,None]

    v = b0 - np.sum(b0*b1, axis=-1)[:,None]*b1
    w = b2 - np.sum(b2*b1, axis=-1)[:,None]*b1

    x = np.sum(v*w, axis=-1)
    y = np.sum(np.cross(b1, v)*w, axis=-1)

    return np.arctan2(y, x)

# calculate planar angles defined by 3 sets of points
def get_angles(a, b, c):

    v = a - b
    v /= np.linalg.norm(v, axis=-1)[:,None]

    w = c - b
    w /= np.linalg.norm(w, axis=-1)[:,None]

    x = np.sum(v*w, axis=1)

    return np.arccos(x)

# get 6d coordinates from x,y,z coords of N,Ca,C atoms
def get_coords6d(xyz, dmax):

    nres = xyz.shape[1]

    # three anchor atoms
    N  = xyz[0]
    Ca = xyz[1]
    C  = xyz[2]

    # recreate Cb given N,Ca,C
    b = Ca - N
    c = C - Ca
    a = np.cross(b, c)
    Cb = -0.58273431*a + 0.56802827*b - 0.54067466*c + Ca

    # fast neighbors search to collect all
    # Cb-Cb pairs within dmax
    kdCb = scipy.spatial.cKDTree(Cb)
    indices = kdCb.query_ball_tree(kdCb, dmax)

    # indices of contacting residues
    idx = np.array([[i,j] for i in range(len(indices)) for j in indices[i] if i != j]).T
    idx0 = idx[0]
    idx1 = idx[1]

    # Cb-Cb distance matrix
    dist6d = np.full((nres, nres),999.9)
    dist6d[idx0,idx1] = np.linalg.norm(Cb[idx1]-Cb[idx0], axis=-1)

    # matrix of Ca-Cb-Cb-Ca dihedrals
    omega6d = np.zeros((nres, nres))
    omega6d[idx0,idx1] = get_dihedrals(Ca[idx0], Cb[idx0], Cb[idx1], Ca[idx1])

    # matrix of polar coord theta
    theta6d = np.zeros((nres, nres))
    theta6d[idx0,idx1] = get_dihedrals(N[idx0], Ca[idx0], Cb[idx0], Cb[idx1])

    # matrix of polar coord phi
    phi6d = np.zeros((nres, nres))
    phi6d[idx0,idx1] = get_angles(Ca[idx0], Cb[idx0], Cb[idx1])

    return dist6d, omega6d, theta6d, phi6d

################################
to1letter = {
    "ALA":'A', "ARG":'R', "ASN":'N', "ASP":'D', "CYS":'C',
    "GLN":'Q', "GLU":'E', "GLY":'G', "HIS":'H', "ILE":'I',
    "LEU":'L', "LYS":'K', "MET":'M', "PHE":'F', "PRO":'P',
    "SER":'S', "THR":'T', "TRP":'W', "TYR":'Y', "VAL":'V' }


# read and extract xyz coords of N,Ca,C atoms
# from a PDB file
def parse_pdb(filename):

    lines = open(filename,'r').readlines()

    N  = np.array([[float(l[30:38]), float(l[38:46]), float(l[46:54])]
                    for l in lines if l[:4]=="ATOM" and l[12:16].strip()=="N"])
    Ca = np.array([[float(l[30:38]), float(l[38:46]), float(l[46:54])]
                    for l in lines if l[:4]=="ATOM" and l[12:16].strip()=="CA"])
    C  = np.array([[float(l[30:38]), float(l[38:46]), float(l[46:54])]
                    for l in lines if l[:4]=="ATOM" and l[12:16].strip()=="C"])

    xyz = np.stack([N,Ca,C], axis=0)

    # indices of residues observed in the structure
    idx = np.array([int(l[22:26]) for l in lines if l[:4]=="ATOM" and l[12:16].strip()=="CA"])

    return xyz,idx

def loader(pdb_path, fasta_path, params):
    xyz,idx = parse_pdb(pdb_path)
    seq = parse_fasta(fasta_path)
    
    # get 6D coords
    d,o,t,p = get_coords6d(xyz, params['DMAX'])

    # bin 6D coords
    dstep = (params['DMAX'] - params['DMIN']) / params['DBINS']
    dbins = np.linspace(params['DMIN']+dstep, params['DMAX'], params['DBINS'])

    astep = 2.0*np.pi / params['ABINS']
    ab360 = np.linspace(-np.pi+astep, np.pi, params['ABINS'])
    ab180 = np.linspace(astep, np.pi, int(params['ABINS']/2))

    db = np.digitize(d, dbins).astype(np.uint8)
    ob = np.digitize(o, ab360).astype(np.uint8)
    tb = np.digitize(t, ab360).astype(np.uint8)
    pb = np.digitize(p, ab180).astype(np.uint8)

    # synchronize 'no contact' bins
    ob[db == params['DBINS']] = params['ABINS']
    tb[db == params['DBINS']] = params['ABINS']
    pb[db == params['DBINS']] = int(params['ABINS']/2)

    # stack all coords together
    c6d = np.stack([db,ob,tb,pb], axis=0)

    # slice long chains
    L = idx.shape[0]
    start,stop,nres = 0,L,L
    sel = np.arange(0, L)
    if L > params['LMAX']:

        if params['SLICE'] == 'CONT':
            # slice continuously
            nres = np.random.randint(params['LMIN'],params['LMAX'])
#            nres = params['LMAX']
            start = np.random.randint(L-nres+1)
#            start = 0
            stop = start + nres
            sel = np.arange(start,stop)
    
    if (idx[sel][1:]-idx[sel][:-1]==1).all(): #check if idx has no gaps
        feed_dict = {
        "idx"      : idx[sel],
        "idx_raw"  : idx,
        "coords6d" : c6d[:,sel,:][:,:,sel],
        "sel"      : sel,
        "dist"     : d[sel,:][:,sel],
        "fullseq"  : [seq[0]]
        }
    else:
        feed_dict = {
        "idx"      : idx[sel],
        "idx_raw"  : idx,
        "coords6d" : np.zeros(1), 
        "sel"      : sel,
        "dist"     : d[sel,:][:,sel],
        "fullseq"  : [seq[0]]
        }    
    return feed_dict