import torch
import torch.nn as nn
import torch.nn.functional as F

import matplotlib.pyplot as plt

class LM_scorer(nn.Module):
    """
    Calculate -log(p) of a sequence (per residue and mean)
    """
    def __init__(self, lm_model, alphabet, device='cpu', bkg_freq_dict=None):
        super(LM_scorer, self).__init__()
        self.LM_alphabet = ['L', 'A', 'G', 'V', 'S', 'E', 'R', 'T', 'I', 'D', 'P', 'K', 'Q', 'N', 'F', 'Y', 'M', 'H', 'W', 'C']
        self.device = device
        self.lm_model = lm_model.to(self.device)
        self.batch_converter = alphabet.get_batch_converter()
        self.alphabet = alphabet
        self.lm_model.eval()
        self.bkg_freq_dict = bkg_freq_dict  # dict of AA1: bkg_frequency
        
        
    def get_masked_tokens(self, seq=None, tokens_one_hot=None):
        """This function maps seq, or tokens_one_hot to tokens_one_hot_masked. It is for a single example, i.e. batch=1
        Args:
          seq (str)
          tokens_one_hot (torch.float32): [sequence_length, 33]
        Returns:
          tokens_one_hot_masked (torch.float32): [sequence_length-2, sequence_length, 33]
        """        
        if tokens_one_hot is not None:
            seq_len = tokens_one_hot.shape[0]
            bool_mask = torch.zeros([seq_len-2, seq_len]).long().to(self.device)
            bool_mask[:,1:-1] = torch.eye(seq_len-2).to(self.device)
            masked_batch_tokens = tokens_one_hot.repeat([seq_len-2, 1, 1])
            tokens_one_hot_masked = masked_batch_tokens*(1-bool_mask[:,:,None])+torch.eye(33).to(self.device)[self.alphabet.mask_idx]*bool_mask[:,:,None]
        elif seq is not None:
            masked_seqs = [("name", [j if i!=num else "<mask>"  for i, j in enumerate(seq)]) for num in range(len(seq))]
            batch_labels, batch_strs, masked_batch_tokens = self.batch_converter(masked_seqs)
            tokens_one_hot_masked = torch.eye(33).to(self.device)[masked_batch_tokens]
        return tokens_one_hot_masked


    def batch(self, iterable, n=1):
        l = len(iterable)
        for ndx in range(0, l, n):
            yield iterable[ndx:min(ndx + n, l)]


    def get_sequence_logits(self, batch_tokens=None, batch_tokens_one_hot=None):
        """This function maps batch_tokens, or batch_tokens_one_hot to sequence logits.
        Args:
          batch_tokens (torch.int32): [batch_size, sequence_length]
          batch_tokens_one_hot (torch.float32): [batch_size, sequence_length, 33] 
        Returns:
          logits (torch.float32): [batch_size, sequence_length, 33]
        """

        if batch_tokens_one_hot is not None:
            batch_tokens = torch.argmax(batch_tokens_one_hot, axis=-1) #these tokens are not used in the language model
            out = self.lm_model(batch_tokens, batch_tokens_one_hot, return_contacts=False)
        elif batch_tokens is not None:
            out = self.lm_model(batch_tokens, return_contacts=False)
        return out['logits']


    def get_SSM(self, tokens_one_hot_masked, batch_size):
        """This function maps tokens_one_hot_masked, batch_size to sequence probabilities.
        Args:
          tokens_one_hot_masked (torch.float32): [sequence_length-2, sequence_length, 33]
          batch_size (int): batch size for passing to the Language Model
        Returns:
          seq_probs (torch.float32): [sequence_length-2, 20]
        """
        B, L, _ = tokens_one_hot_masked.shape
        logits_list = []
        for i in self.batch(range(0, B), batch_size):
            logits = self.get_sequence_logits(batch_tokens_one_hot=tokens_one_hot_masked[i,:,:])
            logits_list.append(logits)

        logits = torch.cat(logits_list)[:,1:-1,4:24]  # get top 20 AAs 
        seq_probs = (torch.diagonal(torch.softmax(logits, axis=-1))).transpose(1,0)
        return seq_probs

    
    def alphabet_onehot_2_onehot(self, alphabet1, alphabet2):
        '''
        Args:
        alphabet1: List of amino acids in order (A characters). Len A1
        alphabet2: List of amino acids in different order. Len A2

        Returns:
        map: A2xA1 matrix to map one-hot encoding from alphabet1 to alphabet2
        '''
        assert len(alphabet1) <= len(alphabet2), 'Alphabet 1 must be smaller than or equal to alphabet 2'

        alpha1_2_int = {aa: i for i, aa in enumerate(alphabet1)}
        alpha2_2_int = {aa: i for i, aa in enumerate(alphabet2)}

        A1 = len(alphabet1)
        A2 = len(alphabet2)
        map_ = torch.zeros((A2, A1), dtype=torch.float32).to(self.device)

        for aa in alphabet1:
            j = alpha1_2_int[aa]
            i = alpha2_2_int[aa]
            map_[i, j] = 1

        return map_
    
    
    def get_sequence_cce(self, seq_str=None, seq_oh=None, encoding_alphabet=None, batch_size=None, eps=1e-12):
        '''
        seq_str: string of AA1. Does not allow gradient backprop
        seq_oh: One-hot encoding of the sequence. If it was made with gumble straight through trick,
                gradient can be passed backwards. shape = (L,A). L is effectively the batch size.
        encoding_alphabet: alphabet used to encode seq_oh. List of 20 amino acids (single letters).
        pssm: probability distribution of each AA per position. If provided, it will be used as the 
                "true" prob distribution instead of a one-hot distribution when calculating cross entropy.
        '''
        assert (seq_str is None) ^ (seq_oh is None), 'Either seq_str or seq_oh should be passed, not both'  # ^ = XOR
        
        if seq_oh is not None:
            # Map to convert seq one-hot from encoding alphabet to LM alphabet
            map_alpha_2_lm = self.alphabet_onehot_2_onehot(encoding_alphabet, self.alphabet.all_toks)
            
            # Permute dim=1 of seq_oh to match LM sequence encoding. (CAN PROBABLY REPLACE ALL 3 LINES WITH EINSUM)
            seq_oh_tok = seq_oh.permute(1,0)
            seq_oh_tok = map_alpha_2_lm @ seq_oh_tok
            seq_oh_tok = seq_oh_tok.permute(1,0)
            
            # Force leading <cls>, Met and trailing <eos>
            seq_oh_tok = torch.cat([torch.zeros((2,33)).to(self.device), seq_oh_tok, torch.zeros((1,33)).to(self.device)], axis=0)
            seq_oh_tok[0,0] = 1.  # cls
            seq_oh_tok[1,20] = 1.  # Met
            seq_oh_tok[-1,2] = 1.  # eos
            
            # Sanity check. Str representation of seq_oh
            seq_str = [self.alphabet.all_toks[idx] for idx in seq_oh_tok.argmax(1)]
            seq_str = ''.join(seq_str)
            #print('LM_scorer read one-hot sequence as:', seq_str)
            
            # Batch of masked tokens
            tokens_one_hot_masked = self.get_masked_tokens(tokens_one_hot=seq_oh_tok)
            
            # make p_true for cce calculation
            p_true = seq_oh_tok[1:-1, 4:24]  # shape=(L, 20). includes leading met
            
        elif seq_str is not None:
            # Prepend Met
            seq_str = 'M' + seq_str
        
            # Tokenize and get masked positions
            data_ = [("protein", seq_str)]
            _, _, sequence_tokens = self.batch_converter(data_)
            tokens_one_hot_masked = self.get_masked_tokens(seq_str)
            
            # make p_true for cce calculation
            p_true = F.one_hot(sequence_tokens[0,1:-1],33)[:,4:24]
            p_true = p_true.to(self.device)  # for mysterious reasons, this runs MUCH faster if done before the LM forward pass
        
        # Load bkg aa freq. IT WOULD BE NICE TO DO THIS AFTER SSM CALC WITHOUT TAKING ~2 SECONDS!
        if self.bkg_freq_dict is not None:
            bkg_freq = [self.bkg_freq_dict[aa] for aa in self.LM_alphabet]
            bkg_freq = torch.tensor(bkg_freq).to(self.device)[None]  # (1, A)
            bkg_by_res = -(p_true * torch.log(bkg_freq + eps)).sum(1)  # (L,)
            bkg_by_res = bkg_by_res[1:]  # cut off leading M
           
        # LM predictions
        L_seq = tokens_one_hot_masked.shape[0]
        lm_probs = self.get_SSM(tokens_one_hot_masked, batch_size=L_seq)  # (L,20)

        # Cut off leading Met
        lm_probs = lm_probs[1:]
        p_true = p_true[1:]
        
        # Cross entropy
        ce_by_res = -(p_true * torch.log(lm_probs + eps)).sum(1)
        ce_mean = ce_by_res.mean()
        
        # Output dict
        out = {
            'ce_mean': ce_mean,
            'ce_by_res': ce_by_res,
        }
        
        # bkg correction
        if self.bkg_freq_dict is not None:            
            ce_by_res_bkg_correction = ce_by_res - bkg_by_res
            ce_mean_bkg_correction = ce_by_res_bkg_correction.mean()
            out.update({
                'ce_by_res_bkg_correction': ce_by_res_bkg_correction,
                'ce_mean_bkg_correction': ce_mean_bkg_correction,
            })
            
        return out
    
    
    def forward(self, seq_str=None, seq_oh=None, encoding_alphabet=None, batch_size=None, eps=1e-12):
        """
        seq_str: str of AA1
        seq_oh: One-hot representation of sequence using the encoding alphabet. Pass a gumble straight through approximation
                if you want to packpropogate gradients.
        encoding_alphabet: alphabet used to encode seq_oh. List of 20 amino acids (single letters).
        """
        out = self.get_sequence_cce(seq_str=seq_str, seq_oh=seq_oh, encoding_alphabet=encoding_alphabet, batch_size=batch_size)
        
        return out
