#!/usr/bin/env python3

# calculate the sum of -log(p) (CCE) of a sequence according to the ESM language model

import sys, os
import torch
import torch.nn as nn
import torch.nn.functional as F
import argparse
import pandas as pd
import esm
import lm_scorer

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

LM_MODEL_PATH = '/projects/ml/language_models/models/ESM_models/esm1b_t33_650M_UR50S.pt'
LM_BKG = 'LM_bkg_aa.txt'


def main():
    # parse args
    p = argparse.ArgumentParser()
    p.add_argument('--fas', nargs='+', type=str, help='FASTA file of protein sequence to score.')
    p.add_argument('--separate_output_files', type=bool, default=False, help='Save scores in separate .csv files for each .fas input?')
    p.add_argument('--out_dir', help='Output directory. Defaults to same dir as .fas file.')
    args = p.parse_args()

    # load bkg aa
    with open('language_model/LM_bkg_aa.txt', 'r') as f_in:
        bkg_dict = eval(f_in.read())
    
    # load LM
    print('Loading language model...')
    LM_model, alphabet = esm.pretrained.load_model_and_alphabet(LM_MODEL_PATH)
    LM_model.to(device)
    for p in LM_model.parameters(): p.requires_grad = False
    scorer = lm_scorer.LM_scorer(LM_model.to(device), alphabet, bkg_freq_dict=bkg_dict)
    
    # initialize data holders
    index = []
    data = {
        'ce_mean': [],
        'ce_by_res': [],
        'ce_mean_bkg_correction': [],
        'ce_by_res_bkg_correction': [],
    }
    
    # loop over fasta file(s)
    print('Scoring files...')
    for i, f_fas in enumerate(args.fas):
        if i%100 == 0:
            print(f'Scoring design {i}...')
      
        bn = f_fas.split('/')[-1].split('.')[0]
        
        # read fas file
        with open(f_fas, 'r') as f_in:
            name = f_in.readline().replace('>', '').strip()
            seq = f_in.readline().strip()
            
        # score with LM
        out = scorer.get_sequence_cce(seq_str=seq)
        out = {k: v.cpu().tolist() for k,v in out.items()}

        # update data holders
        index.append(name)
        for k,v in data.items():
            v.append(out[k])

        # write individual .csv files
        if args.separate_output_files:
            # find out dir
            if args.out_dir is None:
                out_dir = os.path.dirname(f_fas)
            else:
                out_dir = args.out_dir

            # save
            df = pd.DataFrame(index=index, data=data)
            df.to_csv(f'{out_dir}/{bn}_lm_score.csv', sep=';')
            
            # reset data holders
            index = []
            data = {}
            
    # write one big result file
    if not args.separate_output_files:
        # find out dir
        if args.out_dir is None:
            out_dir = '/'+os.path.join(*os.path.abspath(f_fas).split('/')[:-1])
        else:
            out_dir = args.out_dir
            
        # save
        df = pd.DataFrame(index=index, data=data)
        df.to_csv(f'{out_dir}/lm_scores.csv', sep=';')
        
        
if __name__ == '__main__':
    main()
            
            
            
            
