import sys, os, pickle, json, importlib, copy
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

#script_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
#include_dir = script_dir+'/main_smooth'
#
#sys.path.insert(0, script_dir+'/main_smooth/')
#sys.path.insert(0, script_dir+'/main_smooth/util')
from main_smooth.trfold.trFold import TRFold
from main_smooth.util.util import num2aa, write_pdb, aa_1_N, aa_N_1, alphabet_mapping, alpha_1, alphabet_onehot_2_onehot 

def get_c6d_dict(out, grad=True):
    C6D_KEYS = ['dist','omega','theta','phi']
    if grad:
        logits = [out[key] for key in C6D_KEYS]
    else:
        logits = [out[key].detach() for key in C6D_KEYS]
    probs = [F.softmax(l, dim=1) for l in logits]
    dict_pred = {}
    dict_pred['p_dist'] = probs[0].permute([0,2,3,1])
    dict_pred['p_omega'] = probs[1].permute([0,2,3,1])
    dict_pred['p_theta'] = probs[2].permute([0,2,3,1])
    dict_pred['p_phi'] = probs[3].permute([0,2,3,1])
    return dict_pred, probs

def gumbel_st_sample(logits):
    """This function takes logits, constructs softmax and uses Gumbel straight-through to sample the argmax.
    Args:
      logits (torch.float32): [batch_size, num_sequences, sequence_length, 20]
    Returns:
      msa (torch.float32): [batch_size, num_sequences, sequence_length]
    """
    device = logits.device
    B, N, L, _ = logits.shape

    LM_y_logprobs = (logits).log_softmax(-1)
    LM_y_probs = LM_y_logprobs.exp()    
    LM_y_seq = F.one_hot(torch.argmax(LM_y_probs, -1), 21).float() #argmax
    LM_y_seq = (LM_y_seq-LM_y_probs).detach() + LM_y_probs #gumbel trick
    return LM_y_seq

class NSGD(torch.optim.Optimizer):
  def __init__(self, params, lr, dim):
    defaults = dict(lr=lr)
    super(NSGD, self).__init__(params, defaults)
    self.dim=dim
  @torch.no_grad()
  def step(self, closure=None):
    loss = None
    if closure is not None:
      with torch.enable_grad():
        loss = closure()
    for group in self.param_groups:
      for p in group['params']:
        if p.grad is None: continue
        d_p = p.grad / (torch.norm(p.grad, dim=self.dim, keepdim=True) + 1e-8)
        p.add_(d_p, alpha=-group['lr'])
    return loss
        
def run_trfold(out, probs, args):
    # TRFold
    with torch.enable_grad():
        probs = [p[0] for p in probs]  # Cut batch dimension
        TRF = TRFold(probs)
        if 'xyz' in out.keys(): 
            xyz = TRF.fold(ca_trace=out['xyz'][0,:,1],
                           batch=args.batch,lr=args.lr,nsteps=args.nsteps)
        elif 'lognorm' in out.keys():
            xyz = TRF.fold(lognorm=out['lognorm'],
                           batch=args.batch,lr=args.lr,nsteps=args.nsteps)
        else:
            sys.exit('TRFold cannot generate an initial structure.')
    return xyz
 

def save_result(out_prefix, Net, ml, msa, args, trb, trk):

    B,N,L = msa.shape
    trb = copy.copy(trb)
    
    device = next(Net.parameters()).device
    Net.eval()
    with torch.no_grad():
        out = Net(torch.tensor(msa).long().to(device))
        dict_pred, probs = get_c6d_dict(out, grad=False)
        E_0 = ml.score({'c6d': dict_pred, 'xyz': out['xyz']})
        E_0 = E_0.cpu().numpy()
        
    # sort designs by loss if batch>1. batch 0 is the best design
    if B > 1:
        idx = np.argsort(E_0)
        E_0 = E_0[idx]
        msa = msa[idx]
    
    trb['loss_tot'] = E_0.tolist()
    for name, value in ml.losses.items():
        trb['loss_'+name] = value.cpu().tolist()

    msa_list_NN = msa.reshape(-1,msa.shape[-1]).tolist()
    msa_list_AA = alphabet_mapping(msa_list_NN, aa_N_1)
    designed_seqs = ["".join(msa_list_AA[i]) for i in range(B)]

    trb['seq'] = designed_seqs[0]

    np.savez(out_prefix+'.npz',
        dist=dict_pred['p_dist'].detach().cpu().numpy()[0,],
        omega=dict_pred['p_omega'].detach().cpu().numpy()[0,],
        theta=dict_pred['p_theta'].detach().cpu().numpy()[0,],
        phi=dict_pred['p_phi'].detach().cpu().numpy()[0,],
        mask=trb['mask_contig'])

    # save sequence from best batch
    with open(out_prefix+'.fas', "w") as f_out:
        print(">" + os.path.basename(out_prefix), file=f_out)
        print(designed_seqs[0], file=f_out)
            
    # save sequence from all batches
    if args.save_batch_fas:
        with open(out_prefix+'_batch.fas', "w") as f_out:
            for i in range(B):
                print(">" + os.path.basename(out_prefix+f"_b{i}"), file=f_out)
                print(designed_seqs[i], file=f_out)

    # Save tracker
    with open(out_prefix+'.trb', 'wb') as f_out:
        pickle.dump(trb, f_out)

    if args.track_step is not None:
        with open(out_prefix+'.trk', 'wb') as f_out:
            pickle.dump(trk, f_out)

    with open(args.out+'.txt', 'a') as f_out:
        loss_tot_str = ' '.join(map(str,trb["loss_tot"]))
        loss_cce_str = ' '.join(map(str,trb["loss_cce"]))
        loss_entropy_str = ' '.join(map(str,trb["loss_entropy"]))
        sampled_mask = trb["sampled_mask"]
        mask = args.mask
        contigs = args.contigs

        print(os.path.basename(out_prefix)+f' {designed_seqs[0]} mask: {mask} contigs: {contigs} sampled_mask: {sampled_mask} loss: {loss_tot_str} cce: {loss_cce_str} entropy: {loss_entropy_str}', file=f_out)

    # Save pdb
    if args.save_pdb:
        print('Generating a trFolded backbone pdb...')
        xyz = run_trfold(out, probs, args)
       
        # write file
        comments = ['##### Meta data #####']
        comments += [f'{k}: {v}' for k, v in args.__dict__.items()]
        comments += [f'{k}: {trb[k]}' for k in ['con_ref_idx0',
                                                'con_hal_idx0',
                                                'con_ref_pdb_idx',
                                                'con_hal_pdb_idx',
                                                'sampled_mask']]

        #comments += [f'{k}: {v}' for k, v in mappings.items()]
        
        write_pdb(xyz=xyz.permute(1,0,2).detach().cpu().numpy(), 
                  prefix=out_prefix,
                  res=designed_seqs[0],
                  comments=comments)             

def load_structure_predictor(include_dir, args, device):
    def import_method(path_str):
        '''import a method from folder/subfolder/file.py given a string
        folder.subfolder.file.method; add folder/subfolder to sys.path'''
        module,method = path_str.rsplit('.',1)
        sys.path.insert(0, include_dir+'/'+module.rsplit('.',1)[0].replace('.','/'))
        module = importlib.import_module(module)
        method = getattr(module,method)
        return method

    print(f'Loading structure prediction model onto device {device}...')
    reg_models = json.load(open(include_dir+"/params/models.json"))
    torch.set_num_threads(4)
    for k,v in reg_models.items():
        flag = '*' if k==args.network_name else ' '
        print("# %c %-15s   [ens=%d]   %s"%(flag,k,len(v[-1]),v[0]))

    sel = reg_models[args.network_name]

    if not os.path.exists(sel[2]):
        sel[2] = '/projects/ml/'+sel[2]

    chks = [sel[2]+'/'+chk for chk in sel[4]]
    NetClass = import_method(sel[1])

    net_params = json.load(open(sel[2]+'/'+sel[3]))

    # command-line modifications to model params
    if 'trr2' in args.network_name:
        net_params['dropout'] = args.drop
    elif 'trunk' in args.network_name:
        net_params['p_drop'] = args.drop

    print(f'\nModel parameters:\n{net_params}')

    if len(chks)>1:
        Net = EnsembleNet(NetClass, net_params, chks)
    else:
        Net = NetClass(**net_params)
        weights = torch.load(chks[0],map_location=torch.device(device))
        if 'model_state_dict' in weights.keys():
            weights = weights['model_state_dict']
        Net.load_state_dict(weights, strict=False)

    Net = Net.to(device)
    Net.eval()
    for p in Net.parameters(): p.requires_grad = False        
    
    return Net, net_params
 

def initialize_logits(args, mappings, L, device, B=1):

    N = args.msa_num

    if args.spike == 0:
        input_logits = args.init_sd*torch.randn([B,N,L,21]).to(device).float()
        input_logits[...,20] = -1e9 # avoid gaps
    else:
        #1. scatter ref contig sequence to hal contigs
        seq_init = torch.zeros([B,1,L], dtype=torch.int64)
        seq_ref = torch.tensor(pdb['seq'], dtype=torch.int64)
        seq_init[:,:, mappings['con_hal_idx0']] = seq_ref[mappings['con_ref_idx0']]
        n_cat = torch.tensor(20)
        seq_init = F.one_hot(seq_init, n_cat).float()  # (B,1,L,20)
        
        #2. weight logits so they are sampled at desired frequency
        p = torch.tensor(args.spike, dtype=torch.float32)
        g = torch.log(p) - torch.log(1 - p) + torch.log(n_cat.float() - 1.)  # desired gap in logit magnitude
        seq_init *= g
    
        #3. mask out non-contig regions
        seq_init *= torch.tensor(mappings['mask_1d'][0])[None,None,:,None]
        
        #4. Extend to an MSA  (spiking only really makes sense if 1 seqeunce in the MSA)
        input_logits = torch.tensor( #[N,L,20]
            torch.cat([torch.cat([seq_init, 
                                  -10.0*torch.ones([B,1,L,1])],  # make gap very unlikely
                                 axis=-1),
                       torch.zeros([B,N-1,L,21])],  # rest of msa starts at 0
                      axis=1) 
            + args.init_sd*torch.randn([B,N,L,21]),  # add a little noise
            dtype=torch.float32,
            device=device,
            requires_grad=True
        )

    #if args.exclude_aa is not None:
    #    aa_n = [aa_1_N[aa] for aa in args.exclude_aa]
    #    mask_aa = torch.zeros((1,1,21), requires_grad=False)
    #    mask_aa[0, 0, aa_n] = -1e9
    #    input_logits = input_logits + mask_aa.to(device)

    # copy and turn on gradients
    input_logits = torch.tensor(input_logits).to(device).requires_grad_(True)

    return input_logits


def enable_dropout(Net):
    # turn on dropout
    for m in Net.modules():
      if m.__class__.__name__.startswith('Dropout'):
        m.train()
    return Net
            
def gradient_descent(Net, ml, input_logits, args, trb, trk, force_aa=None, exclude_aa=None):

    device = next(Net.parameters()).device

    print('Starting gradient descent...')
    print(f'{"step":>12}',end='')
    ml.print_header_line()

    torch.set_grad_enabled(True)
    if args.drop > 0: Net = enable_dropout(Net)

    L = input_logits.shape[2]

    if args.optimizer == 'nsgd':
        optimizer = NSGD([input_logits], lr=args.learning_rate*np.sqrt(L), dim=[-1,-2])
    elif args.optimizer == 'adam':
        optimizer = torch.optim.Adam([input_logits], lr=args.learning_rate)

    msa = torch.argmax(input_logits, -1).detach().cpu().numpy()

    for step in range(args.grad_steps+1): # 1 extra iteration to compute final outputs
        optimizer.zero_grad()

        if step == args.grad_steps: # no update on last iteration
            torch.set_grad_enabled(False)

        # force and exlucde AAs
        input_logits_biased = force_msa(input_logits, 
                                        alphabet=aa_1_N, 
                                        force_aa=force_aa, 
                                        exclude_aa=exclude_aa)
        
        # sampling trick
        msa_one_hot = gumbel_st_sample(input_logits_biased)

        kwargs = {}
        if args.grad_check and 'trunk' in args.network_name:
            kwargs = dict(use_transf_checkpoint=True)

        out = Net(torch.argmax(msa_one_hot,axis=-1).to(device), msa_one_hot=msa_one_hot.to(device),
                  **kwargs)
        dict_pred, probs = get_c6d_dict(out)

        # Scoring with MultiLoss wrapper!
        net_out = {'c6d': dict_pred, 'xyz': out.get('xyz', None), 'msa_one_hot': msa_one_hot}

        E_0 = ml.score(net_out)
         
        # track intermediate losses
        if args.track_step is not None and step % args.track_step == 0:
            trk['step'].append(int(step))
            trk['step_type'].append('grad')
            trk['loss_tot'].append(float(E_0))
            for name, value in ml.losses.items():
                trk['loss_'+name].append(float(value))

        # best design so far
        if E_0 < trb['loss_tot']:
            trb['loss_tot'] = float(E_0)
            msa = torch.argmax(msa_one_hot.detach(),axis=-1).cpu().numpy()  #(B,N,L), B=1, N=1
            trb['msa'] = msa
       
        if step % 10 == 0:
            print(f'{step:>12}', end='')
            ml.print_losses_line()

        # output intermediate result
        if args.out_step is not None and step > 0 and step % args.out_step == 0:
            save_result(f'{trb["out_prefix"]}_g{step}', Net, ml, msa, args, trb, trk)
            if args.drop > 0: Net = enable_dropout(Net) # dropout gets turned off when saving result

        if step != args.grad_steps: # no update on last iteration
            E_0.backward()
            optimizer.step()

    return msa, out
    

def mcmc(Net, sampler, ml, msa, args, trb, trk, mappings=None, residues_to_force=None, B_mcmc=4):

    Net.eval() # turn off dropout for MCMC
    torch.set_grad_enabled(False)
    device = next(Net.parameters()).device
    
    # Allow batches larger than 1
    B, N, L = msa.shape    
    if (B == 1) and (N == 1):
        msa = np.tile(msa, (B_mcmc,1,1))
        B, N, L = msa.shape

    # Prevent LM sampler from mutating forced positions
    do_not_mutate = []
    if args.force_aa is not None:
        pdbref_2_halidx0 = dict(zip(mappings['con_ref_pdb_idx'], mappings['con_hal_idx0']))
        
        for pdbref, aa in residues_to_force.items():
            idx0 = pdbref_2_halidx0[pdbref]
            idx1 = idx0 + 1
            aa_num = aa_1_N[aa.upper()]
            msa[0, 0, idx0] = aa_num
            
            do_not_mutate.append(idx1)   
    
    print('Starting MCMC...')
    print(f'{"step":>12}',end='')
    ml.print_header_line()

    sf_0 = np.exp(np.log(0.5) / args.mcmc_halflife)  # scaling factor / step

    # initial score
    out = Net(torch.tensor(msa).long().to(device))
    dict_pred, probs = get_c6d_dict(out, grad=False)
    net_out = {'c6d': dict_pred, 'xyz': out.get('xyz', None)}
    E_0 = ml.score(net_out)

    for step in range(args.mcmc_steps):
        sf = sf_0 ** step
        msa_list_NN = msa.reshape(-1,msa.shape[-1]).tolist()
        msa_list_AA = alphabet_mapping(msa_list_NN, aa_N_1)
        msa_list_AA_mutated = sampler(msa_list_AA, 1, sf * args.T_lm_0, do_not_mutate=do_not_mutate, exclude_aa=args.exclude_aa)
        msa_list_NN_mutated = alphabet_mapping(msa_list_AA_mutated, aa_1_N)
        mutated_msa = np.array(msa_list_NN_mutated).reshape(B, N, L)

        out = Net(torch.tensor(mutated_msa).long().to(device))
        dict_pred, probs = get_c6d_dict(out, grad=False)
        
        # Scoring with MultiLoss wrapper!
        net_out = {'c6d': dict_pred, 'xyz': out.get('xyz', None)}
        E_1 = ml.score(net_out)
        
        # Metropolis criteria for each batch 
        T_acc = sf * args.T_acc_0
        acceptances = torch.clamp(torch.exp(-(torch.tensor(E_1-E_0))/T_acc), min=0.0, max=1.0)
        uniform_random = torch.rand(B).to(device)
        bool_accept = (uniform_random < acceptances).long()

        E_0 = bool_accept * E_1 + (1-bool_accept) * E_0 
        bool_accept_np = bool_accept.detach().cpu().numpy()
    #     msa[:,0,:] = bool_accept_np[:,None,None] * mutated_msa[:,0,:] + (1-bool_accept_np[:,None,None]) * msa[:,0,:]
        msa[:,:,:] = bool_accept_np[:,None,None] * mutated_msa[:,:,:] + (1-bool_accept_np[:,None,None]) * msa[:,:,:]
        
        # Every X steps, randomly choose fittest sequences to move forward with
        if (step%200 == 0) and (step != 0):
            # get fresh scores
            out = Net(torch.tensor(msa).long().to(device))
            dict_pred, probs = get_c6d_dict(out, grad=False)
            net_out = {'c6d': dict_pred, 'xyz': out.get('xyz', None)}
            E_1 = ml.score(net_out)
            
            L_ = torch.tensor(L, device=device, dtype=E_1.dtype)
            p = (-E_1 / (T_acc * L_.sqrt())).softmax(0) # normalize sd of E_1
            p = p.cpu().numpy()
            sel = np.random.choice(B,B,p=p)
            msa = msa[sel]
            
            if 'prob' in trk:
                trk['prob'].append(p.tolist())
            else:
                trk['prob'] = [p.tolist()]
        
        # track intermediate losses
        if args.track_step is not None and step % (10*args.track_step) == 0:
            trk['step'].append(int(step))
            trk['step_type'].append('mcmc')
            trk['loss_tot'].append(E_0.cpu().tolist())
            for name, value in ml.losses.items():
                trk['loss_'+name].append(value.cpu().tolist())

        # is there another way to make this work for batch>1?
        if True:  # E_0 < trb['tot_loss']:
            trb['tot_loss'] = E_0.cpu().tolist()
            trb['msa'] = msa.copy()
       
        if step % 100 == 0: 
            print(f'{step:>12}', end='')
            ml.print_losses_line()

        # output intermediate result
        if args.out_step is not None and step > 0 and step % (10*args.out_step) == 0:
            save_result(f'{trb["out_prefix"]}_m{step}', Net, ml, msa, args, trb, trk)

    return msa, out

def force_msa(msa, alphabet, force_aa=None, exclude_aa=None, weight=1e8):
    '''
    Force and exclude certain amino acids in an MSA
    
    Inputs
    -------------
    msa: (torch.tensor, (B,N,L,A))
        Logits of a multiple sequence alignment
    alphabet: (dict)
        Mapping one-letter amino acids (upper case) to int
    exclude_aa: (List)
        List of AA1 to exclude.
    force_aa: (dict)
        key: (int) idx0 of the hallucinated protein.
        val: (str) One letter amino acid
        Forces certain amino acids at certain positions.
        Supersedes exclude_aa
    weight: (float)
        How much to increase/decrease the logit
        
    Outputs
    -------------
    msa: (torch.tensor, (B,N,L,A))
        Updated msa.
    '''
    B,N,L,A = msa.shape
    dtype = msa.dtype
    device = msa.device
    bias = torch.zeros((1,1,L,A), dtype=dtype, device=device)
    
    if exclude_aa is not None:
        for aa in exclude_aa:
            aa = aa.upper()
            bias[0, 0, :, alphabet[aa]] = -1
        
    if force_aa is not None:        
        for idx0, aa in force_aa.items():
            aa = aa.upper()
            bias[0, 0, idx0, alphabet[aa]] = 1
            
    msa = msa + weight * bias
    
    return msa


